@@ -1,62 +1,71 @@
-# -*- coding: utf-8 -*-
 """
 This module contains utilities added by billiard, to keep
 "non-core" functionality out of ``.util``."""
-from __future__ import absolute_import
 
 import os
 import signal
 import sys
 
-import pickle as pypickle
-try:
-    import cPickle as cpickle
-except ImportError:  # pragma: no cover
-    cpickle = None   # noqa
+import pickle
 
 from .exceptions import RestartFreqExceeded
-from .five import monotonic
+from time import monotonic
 
-if sys.version_info < (2, 6):  # pragma: no cover
-    # cPickle does not use absolute_imports
-    pickle = pypickle
-    pickle_load = pypickle.load
-    pickle_loads = pypickle.loads
-else:
-    pickle = cpickle or pypickle
-    pickle_load = pickle.load
-    pickle_loads = pickle.loads
+pickle_load = pickle.load
+pickle_loads = pickle.loads
 
 # cPickle.loads does not support buffer() objects,
 # but we can just create a StringIO and use load.
-if sys.version_info[0] == 3:
-    from io import BytesIO
-else:
+from io import BytesIO
+
+
+SIGMAP = dict(
+    (getattr(signal, n), n) for n in dir(signal) if n.startswith('SIG')
+)
+for _alias_sig in ('SIGHUP', 'SIGABRT'):
     try:
-        from cStringIO import StringIO as BytesIO  # noqa
-    except ImportError:
-        from StringIO import StringIO as BytesIO  # noqa
+        # Alias for deprecated signal overwrites the name we want
+        SIGMAP[getattr(signal, _alias_sig)] = _alias_sig
+    except AttributeError:
+        pass
+
+
+TERM_SIGNAL, TERM_SIGNAME = signal.SIGTERM, 'SIGTERM'
+REMAP_SIGTERM = os.environ.get('REMAP_SIGTERM')
+if REMAP_SIGTERM:
+    TERM_SIGNAL, TERM_SIGNAME = (
+        getattr(signal, REMAP_SIGTERM), REMAP_SIGTERM)
+
+
+TERMSIGS_IGNORE = {'SIGTERM'} if REMAP_SIGTERM else set()
+TERMSIGS_FORCE = {'SIGQUIT'} if REMAP_SIGTERM else set()
 
 EX_SOFTWARE = 70
 
-TERMSIGS = (
+TERMSIGS_DEFAULT = {
+    'SIGHUP',
+    'SIGQUIT',
+    TERM_SIGNAME,
+    'SIGUSR1',
+}
+
+TERMSIGS_FULL = {
     'SIGHUP',
     'SIGQUIT',
     'SIGTRAP',
     'SIGABRT',
     'SIGEMT',
-    'SIGBUS',
     'SIGSYS',
     'SIGPIPE',
     'SIGALRM',
-    'SIGTERM',
+    TERM_SIGNAME,
     'SIGXCPU',
     'SIGXFSZ',
     'SIGVTALRM',
     'SIGPROF',
     'SIGUSR1',
     'SIGUSR2',
-)
+}
 
 #: set by signal handlers just before calling exit.
 #: if this is true after the sighandler returns it means that something
@@ -65,6 +74,15 @@ TERMSIGS = (
 _should_have_exited = [False]
 
 
+def human_status(status):
+    if (status or 0) < 0:
+        try:
+            return 'signal {0} ({1})'.format(-status, SIGMAP[-status])
+        except KeyError:
+            return 'signal {0}'.format(-status)
+    return 'exitcode {0}'.format(status)
+
+
 def pickle_loads(s, load=pickle_load):
     # used to support buffer objects
     return load(BytesIO(s))
@@ -88,19 +106,30 @@ def _shutdown_cleanup(signum, frame):
     sys.exit(-(256 - signum))
 
 
-def reset_signals(handler=_shutdown_cleanup):
-    for sig in TERMSIGS:
-        try:
-            signum = getattr(signal, sig)
-        except AttributeError:
-            pass
-        else:
-            current = signal.getsignal(signum)
-            if current is not None and current != signal.SIG_IGN:
-                maybe_setsignal(signum, handler)
+def signum(sig):
+    return getattr(signal, sig, None)
 
 
-class restart_state(object):
+def _should_override_term_signal(sig, current):
+    return (
+        sig in TERMSIGS_FORCE or
+        (current is not None and current != signal.SIG_IGN)
+    )
+
+
+def reset_signals(handler=_shutdown_cleanup, full=False):
+    for sig in TERMSIGS_FULL if full else TERMSIGS_DEFAULT:
+        num = signum(sig)
+        if num:
+            if _should_override_term_signal(sig, signal.getsignal(num)):
+                maybe_setsignal(num, handler)
+    for sig in TERMSIGS_IGNORE:
+        num = signum(sig)
+        if num:
+            maybe_setsignal(num, signal.SIG_IGN)
+
+
+class restart_state:
     RestartFreqExceeded = RestartFreqExceeded
 
     def __init__(self, maxR, maxT):
